from sklearn.metrics import confusion_matrix
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np

def confusionMatrix(y_pred, y_val):
    """
    Takes Predicted data and Validation data as input and prepares and plots Confusion Matrix.
    """
    try:
        print('''Making Confusion Matrix [*]''')
        cm = confusion_matrix(y_val, y_pred)
        print(cm)
        ax = plt.subplot()
        sns.heatmap(cm, annot=True, fmt='g', ax=ax)
        ax.set_xlabel('Predicted labels')
        ax.set_ylabel('True labels')
        ax.set_title('Confusion Matrix')
        ax.xaxis.set_ticklabels(np.unique(y_val))
        ax.yaxis.set_ticklabels(np.unique(y_val))
        plt.show()
        print('Confusion Matrix Done [', u'\u2713', ']\n')
    except Exception as error:
        print('Building Confusion Matrix Failed with error :', error, '\n')



from sklearn.decomposition import PCA, KernelPCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
import time


def dimensionalityReduction(lda, pca, X_train, X_val, y_train, n_components_lda, n_components_pca, pca_kernel, start):
    """
    Performs Dimensionality Reduction on Training and Validation independent variables.
    """
    try:
        if lda == 'y':
            print('Applying LDA [*]\n')

            lda = LDA(n_components=n_components_lda)
            X_train = lda.fit_transform(X_train, y_train)
            X_val = lda.transform(X_val)
            print('LDA Done [', u'\u2713', ']\n')
        if pca == 'y' and not lda == 'y':
            print('Applying PCA [*]\n')
            if not pca_kernel == 'linear':
                try:

                    kpca = KernelPCA(
                        n_components=n_components_pca, kernel=pca_kernel)
                    X_train = kpca.fit_transform(X_train)
                    X_val = kpca.transform(X_val)
                except MemoryError as error:
                    print(error)
                    end = time.time()
                    print('Time Elapsed :', end - start)
                    return

            elif pca_kernel == 'linear':

                pca = PCA(n_components=n_components_pca)
                X_train = pca.fit_transform(X_train)
                X_val = pca.transform(X_val)
            else:
                print('Un-identified PCA Kernel')
                return
            print('PCA Done [', u'\u2713', ']\n')
        return (X_train, X_val)
    except Exception as error:
        print('Dimensionality Reduction Failed with error :', error, '\n')
        return(X_train, X_val)


from sklearn.model_selection import cross_val_score


def kfold(model, predictor,
          X_train, y_train, cv_folds,isReg
          ):
    """
    Takes predictor, input_units, epochs, batch_size, X_train, y_train, cv_folds, and accuracy_scores dictionary. 
    Performs K-Fold Cross validation and stores result in accuracy_scores dictionary and returns it.
    """
    if not isReg:
        name = {
            'lr' : 'Logistic Regression',
            'svm': 'Support Vector Machine',
            'knn': 'K-Nearest Neighbours',
            'dt' : 'Decision Trees',
            'nb' : 'Naive Bayes',
            'rfc': 'Random Forest Classifier',
            'xgb': 'XGBoost Classifier',
            'ann': 'Artificical Neural Network',
        }
        scoring = 'accuracy'
    if isReg:
        name = {
            'lin' : 'Linear Regression',
            'sgd' : 'Stochastic Gradient Descent Regressor',
            'krr' : 'Kernel Ridge Regressor',
            'elas': 'Elastic Net Regressot',
            'br'  : 'Bayesian Ridge Regressor',
            'svr' : 'Support Vector Regressor',
            'knr' : 'K-Nearest Regressor',
            'dt'  : 'Decision Trees',
            'rfr' : 'Random Forest Regressor',
            'gbr' : 'Gradient Boost Regressor',
            'lgbm': 'LightGB Regressor',
            'xgb' : 'XGBoost Regressor',
            'cat' : 'Catboost Regressor',
            'ann' : 'Artificical Neural Network',
        }
        scoring = 'r2'
    try:
        print('Applying K-Fold Cross validation [*]')
        accuracies = cross_val_score(
            estimator=model, X=X_train, y=y_train, cv=cv_folds, scoring=scoring)
        print("Accuracy: {:.2f} %".format(accuracies.mean()*100))

        model_name = name[predictor]
        accuracy = accuracies.mean()*100

        print("Standard Deviation: {:.2f} %".format(accuracies.std()*100))
        print('K-Fold Cross validation [', u'\u2713', ']\n')
        return (model_name, accuracy)

    except Exception as error:
        print('K-Fold Cross Validation failed with error: ', error, '\n')



import scipy


def sparseCheck(features, labels):
    features = features
    labels = labels
    """
    Takes features and labels as input and checks if any of those is sparse csr_matrix.
    """
    try:
        print('Checking for Sparse Matrix [*]\n')
        if scipy.sparse.issparse(features[()]):
            print('Converting Sparse Features to array []\n')
            features = features[()].toarray()
            print(
                'Conversion of Sparse Features to array Done [', u'\u2713', ']\n')

        elif scipy.sparse.issparse(labels[()]):
            print('Converting Sparse Labels to array []\n')
            labels = labels[()].toarray()
            print(
                'Conversion of Sparse Labels to array Done [', u'\u2713', ']\n')

        else:
            print('No Sparse Matrix Found')

    except Exception as error:
        # print('Sparse matrix Check failed with KeyError: ', error)
        pass
    return (features, labels)



import numpy as np
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
import tensorflow as tf
from luciferml.supervised.utils.classification_params import *


def classificationPredictor(
        predictor, params, X_train, X_val, y_train, y_val, epochs, hidden_layers,
        input_activation, output_activation, loss,
        batch_size, metrics, validation_split, optimizer, output_units, input_units, tune_mode
):
    """
    Takes Predictor string , parameters , Training and Validation set and Returns a classifier for the Choosen Predictor.
    """
    try:
        if predictor == 'lr':
            print('Training Logistic Regression on Training Set [*]\n')

            classifier = LogisticRegression(**params)
            if tune_mode == 1:
                parameters = parameters_lin_1
            elif tune_mode == 2:
                parameters = parameters_lin_2
            elif tune_mode == 3:
                parameters = parameters_lin_3

        elif predictor == 'svm':
            print('Training Support Vector Machine on Training Set [*]\n')

            classifier = SVC(**params)
            if tune_mode == 1:
                parameters = parameters_svm_1
            elif tune_mode == 2:
                parameters = parameters_svm_2
            elif tune_mode == 3:
                parameters = parameters_svm_3

        elif predictor == 'knn':
            print('Training K-Nearest Neighbours on Training Set [*]\n')

            classifier = KNeighborsClassifier(**params)
            if tune_mode == 1:
                parameters = parameters_knn_1
            elif tune_mode == 2:
                parameters = parameters_knn_2
            elif tune_mode == 3:
                parameters = parameters_knn_3

        elif predictor == 'dt':
            print('Training Decision Tree Classifier on Training Set [*]\n')

            classifier = DecisionTreeClassifier(**params)

            if tune_mode == 1:
                parameters = parameters_dt_1
            elif tune_mode == 2:
                parameters = parameters_dt_2
            elif tune_mode == 3:
                parameters = parameters_dt_3

        elif predictor == 'nb':
            print('Training Naive Bayes Classifier on Training Set [*]\n')

            classifier = GaussianNB(**params)
            parameters = {}

        elif predictor == 'rfc':
            print('Training Random Forest Classifier on Training Set [*]\n')

            classifier = RandomForestClassifier(**params)
            if tune_mode == 1:
                parameters = parameters_rfc_1
            elif tune_mode == 2:
                parameters = parameters_rfc_2
            elif tune_mode == 3:
                parameters = parameters_rfc_3

        elif predictor == 'xgb':
            print('Training XGBClassifier on Training Set [*]\n')

            classifier = XGBClassifier(**params)

            if tune_mode == 1:
                parameters = parameters_xgb_1
            elif tune_mode == 2:
                parameters = parameters_xgb_2
            elif tune_mode == 3:
                parameters = parameters_xgb_3
            parameters = {
                'min_child_weight': [1, 5, 10],
                'gamma': [1e-3, 1e-4, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
                'subsample': [0.6, 0.8, 1.0],
                'colsample_bytree': [0.6, 0.8, 1.0],
                'max_depth': [4, 5, 6, 7, 8, 9, 10, 11, 12, 15, 20, 30, 40, 50, 70, 90, 120, 150],
                'learning_rate': [0.3, 0.1, 0.03],
            }
        elif predictor == 'ann':
            def build_ann_model(input_units):
                try:
                    classifier = tf.keras.models.Sequential()
                    for i in range(0, hidden_layers):
                        classifier.add(tf.keras.layers.Dense(
                            units=input_units, activation=input_activation))
                    classifier.add(tf.keras.layers.Dense(
                        units=output_units, activation=output_activation))
                    classifier.compile(optimizer=optimizer,
                                       loss=loss, metrics=metrics)
                    return classifier

                except Exception as error:
                    print('ANN Build Failed with error :', error, '\n')
            print('Training ANN on Training Set [*]\n')
            classifier = build_ann_model(input_units)

            ann_history = classifier.fit(
                X_train, y_train, validation_split=validation_split,
                validation_data=(
                    X_val, y_val), epochs=epochs, batch_size=batch_size
            )
            classifier_wrap = tf.keras.wrappers.scikit_learn.KerasClassifier(
                build_fn=build_ann_model, verbose=1, input_units=input_units,
                epochs=epochs, batch_size=batch_size
            )
            if tune_mode == 1:
                parameters = parameters_ann_1
            elif tune_mode == 2:
                parameters = parameters_ann_2
            elif tune_mode == 3:
                parameters = parameters_ann_3
        if predictor == 'ann':
            return (parameters, classifier,classifier_wrap)
        return (parameters, classifier)
    except Exception as error:
        print('Model Build Failed with error :', error, '\n')



from imblearn.over_sampling import SMOTE
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler


def pred_preprocess(features, labels, test_size, random_state, smote, k_neighbors):

    try:
        if smote == 'y':
            print('Applying SMOTE [*]\n')

            sm = SMOTE(k_neighbors=k_neighbors)
            features, labels = sm.fit_resample(
                features, labels)
            print('SMOTE Done [', u'\u2713', ']\n')

        # Splitting ---------------------------------------------------------------------
        print('Splitting Data into Train and Validation Sets [*]\n')

        X_train, X_val, y_train, y_val = train_test_split(
            features, labels, test_size=test_size, random_state=random_state)
        print('Splitting Done [', u'\u2713', ']\n')

        # Scaling ---------------------------------------------------------------------
        print('Scaling Training and Test Sets [*]\n')

        sc = StandardScaler()
        X_train = sc.fit_transform(X_train)
        X_val = sc.transform(X_val)
        print('Scaling Done [', u'\u2713', ']\n')

        return (X_train, X_val, y_train, y_val)

    except Exception as error:
        print('Preprocessing Failed with error: ', error, '\n')



import time
from typing import Tuple



from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error

import pandas as pd

from luciferml.supervised.utils.encoder import encoder
from luciferml.supervised.utils.predPreprocess import pred_preprocess
from luciferml.supervised.utils.dimensionalityReduction import dimensionalityReduction
from luciferml.supervised.utils.regressionPredictor import regressionPredictor
from luciferml.supervised.utils.confusionMatrix import confusionMatrix
from luciferml.supervised.utils.kfold import kfold
from luciferml.supervised.utils.hyperTune import hyperTune
from luciferml.supervised.utils.sparseCheck import sparseCheck
from luciferml.supervised.utils.intro import intro

class Regression:

    def __init__(self,
                 predictor='lin',
                 params={},
                 tune=False,
                 test_size=.2,
                 cv_folds=10,
                 random_state=42,
                 pca_kernel='linear',
                 n_components_lda=1,
                 lda='n', pca='n',
                 n_components_pca=2,
                 hidden_layers=4,
                 output_units=1,
                 input_units=6,
                 input_activation='relu',
                 optimizer='adam',
                 loss='mean_squared_error',
                 validation_split=.20,
                 epochs=100,
                 batch_size=32,
                 tune_mode=1,
                 smote= 'n',
                 k_neighbors=1
                 ):
        """
        Encodes Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2
        scales X_train, X_val using StandardScaler.
        Fits every model on training set and predicts results,Finds R2 Score and mean square error
        finds accuracy of model applies K-Fold Cross Validation
        and stores its accuracies in a dictionary containing Model name as Key and accuracies as values and returns it
        Applies GridSearch Cross Validation and gives best params out from param list.


        Parameters:
            features : array
                        features array

            lables : array
                        labels array

            predictor : str
                    Predicting model to be used
                    Default 'lin'
                        Predictor Strings:

                                lin  - Linear Regression

                                sgd  - Stochastic Gradient Descent Regressor

                                elas - Elastic Net Regressot

                                krr  - Kernel Ridge Regressor

                                br   - Bayesian Ridge Regressor

                                svr  - Support Vector Regressor

                                knr  - K-Nearest Regressor

                                dt   - Decision Trees

                                rfr  - Random Forest Regressor

                                gbr  - Gradient Boost Regressor

                                lgbm - LightGB Regressor

                                xgb  - XGBoost Regressor

                                cat  - Catboost Regressor

                                ann  - Artificical Neural Network
            params : dict
                        contains parameters for model
            tune : boolean
                    when True Applies GridSearch CrossValidation
                    Default is False

            test_size: float or int, default=.2
                        If float, should be between 0.0 and 1.0 and represent
                        the proportion of the dataset to include in
                        the test split.
                        If int, represents the absolute number of test samples.

            cv_folds : int
                    No. of cross validation folds. Default = 10
            pca : str
                if 'y' will apply PCA on Train and Validation set. Default = 'n'
            lda : str
                if 'y' will apply LDA on Train and Validation set. Default = 'n'
            pca_kernel : str
                    Kernel to be use in PCA. Default = 'linear'
            n_components_lda : int
                    No. of components for LDA. Default = 1
            n_components_pca : int
                    No. of components for PCA. Default = 2
            hidden_layers : int
                    No. of default layers of ann. Default = 4
            inputs_units : int
                    No. of units in input layer. Default = 6
            output_units : int
                    No. of units in output layer. Default = 6
            self.input_activation : str
                    Activation function for Hidden layers. Default = 'relu'
            optimizer: str
                    Optimizer for ann. Default = 'adam'
            loss : str
                    loss method for ann. Default = 'mean_squared_error'
            validation_split : float or int
                    Percentage of validation set splitting in ann. Default = .20
            epochs : int
                    No. of epochs for ann. Default = 100
            batch_size :
                    Batch Size for ANN. Default = 32
            tune_mode : int
                    HyperParam tune modes. Default = 1
                        Available Modes:
                            1 : Basic Tune
                            2 : Intermediate Tune
                            3 : Extreme Tune (Can Take Much Time)
            smote : str,
                Whether to apply SMOTE. Default = 'y'
            k_neighbors : int
                No. of neighbours for SMOTE. Default = 1

        Example:

            from luciferml.supervised import regression as reg

            dataset = pd.read_csv('Social_Network_Ads.csv')

            X = dataset.iloc[:, :-1]

            y = dataset.iloc[:, -1]
            
            reg.Regression(predictor = 'lin').predict(X, y)

        """

        self.predictor = predictor
        self.params = params
        self.tune = tune
        self.test_size = test_size
        self.cv_folds = cv_folds
        self.random_state = random_state
        self.pca_kernel = pca_kernel
        self.n_components_lda = n_components_lda
        self.lda = lda
        self.pca = pca
        self.n_components_pca = n_components_pca
        self.hidden_layers = hidden_layers
        self.output_units = output_units
        self.input_units = input_units
        self.input_activation = input_activation
        self.optimizer = optimizer
        self.loss = loss
        self.validation_split = validation_split
        self.epochs = epochs
        self.batch_size = batch_size
        self.tune_mode = tune_mode
        self.rerun = False
        self.smote = smote
        self.k_neighbors = k_neighbors

        self.accuracy_scores = {}

    def predict(self, features, labels) -> Tuple:
        self.features = features
        self.labels = labels

        # Time Function ---------------------------------------------------------------------

        start = time.time()
        intro()
        print("Started Lucifer-ML \n")
        if not self.rerun:
            # CHECKUP ---------------------------------------------------------------------
            if not isinstance(self.features, pd.DataFrame) and not isinstance(self.labels, pd.Series):
                print('TypeError: This Function take features as Pandas Dataframe and labels as Pandas Series. Please check your implementation.\n')
                end = time.time()
                print(end - start)
                return

            # Encoding ---------------------------------------------------------------------

            self.features, self.labels = encoder(self.features, self.labels)

            # Sparse Check -------------------------------------------------------------
            self.features, self.labels = sparseCheck(self.features, self.labels)

            # Preprocessing ---------------------------------------------------------------------
            self.X_train, X_val, self.y_train, y_val = pred_preprocess(
                self.features, self.labels, self.test_size, self.random_state, self.smote, self.k_neighbors)

            # Dimensionality Reduction---------------------------------------------------------------------
            self.X_train, X_val = dimensionalityReduction(
                self.lda, self.pca, self.X_train, X_val, self.y_train,
                self.n_components_lda, self.n_components_pca, self.pca_kernel, start)

        # Models ---------------------------------------------------------------------
        if self.predictor == 'ann':
            self.parameters, self.regressor, self.regressor_wrap = regressionPredictor(
                self.predictor, self.params, self.X_train, X_val, self.y_train, y_val, self.epochs, self.hidden_layers,
                self.input_activation, self.loss,
                self.batch_size, self.validation_split, self.optimizer, self.output_units, self.input_units, self.tune_mode
            )
        else:
            self.parameters, self.regressor = regressionPredictor(
                self.predictor, self.params, self.X_train, X_val, self.y_train, y_val, self.epochs, self.hidden_layers,
                self.input_activation,  self.loss,
                self.batch_size, self.validation_split, self.optimizer, self.output_units, self.input_units, self.tune_mode
            )

        try:

            if not self.predictor == 'ann':
                self.regressor.fit(self.X_train, self.y_train)
        except Exception as error:
            print('Model Train Failed with error: ', error, '\n')

        print('Model Training Done [', u'\u2713', ']\n')
        print('Predicting Data [*]\n')
        try:
            y_pred = self.regressor.predict(X_val)
            print('Data Prediction Done [', u'\u2713', ']\n')
        except Exception as error:
            print('Prediction Failed with error: ', error,  '\n')


        # Accuracy ---------------------------------------------------------------------
        print('''Evaluating Model Performance [*]''')
        try:
            accuracy = r2_score(y_val, y_pred)
            m_absolute_error = mean_absolute_error(y_val, y_pred)
            m_squared_error = mean_squared_error(y_val, y_pred,squared=False)
            print('Validation R2 Score is {:.2f} %'.format(accuracy*100))
            print('Validation Mean Absolute Error is :',
                  m_absolute_error)
            print('Validation Root Mean Squared Error is :', m_squared_error, '\n')
            print('Evaluating Model Performance [', u'\u2713', ']\n')
        except Exception as error:
            print('Model Evaluation Failed with error: ', error, '\n')

        # K-Fold ---------------------------------------------------------------------
        if self.predictor == 'ann':
            self.regressor_name, accuracy = kfold(
                self.regressor_wrap,
                self.predictor, self.X_train, self.y_train, self.cv_folds,True


            )
        else:
            self.regressor_name, accuracy = kfold(
                self.regressor,
                self.predictor,
                self.X_train, self.y_train, self.cv_folds,True
            )

        # GridSearch ---------------------------------------------------------------------
        if not self.predictor == 'nb' and self.tune:
            self.__tuner()

        print('Complete [', u'\u2713', ']\n')
        end = time.time()
        print('Time Elapsed : ', end - start, 'seconds \n')
        return (self.regressor_name, accuracy)

    def __tuner(self):
        if self.predictor == 'ann':
            self.best_params = hyperTune(
                self.regressor_wrap, self.parameters, self.X_train, self.y_train, self.cv_folds, self.tune_mode,True)
        else:
            self.best_params = hyperTune(
            self.regressor, self.parameters, self.X_train, self.y_train, self.cv_folds, self.tune_mode,True)
        # if self.tune_mode == 3:
        #     self.params = self.best_params
        #     self.tune = False
        #     self.rerun = True
        #     self.predict(self.features, self.labels)
        #     print("Re-ran Predictor on these params :", self.params)
        #     print(
        #         'Re-running regressor with Best Params Done[', u'\u2713', ']\n')

